@@ -45,6 +45,7 @@ ALL_GAMES = ['AirRaid', 'Alien', 'Amidar', 'Assault', 'Asterix', 'Asteroids',
              'SpaceInvaders', 'StarGunner', 'Tennis', 'TimePilot', 'Tutankham',
              'UpNDown', 'Venture', 'VideoPinball', 'WizardOfWor', 'YarsRevenge',
              'Zaxxon']
+MUJOCO_GAMES = ['Ant', 'HalfCheetah', 'Hopper', 'Humanoid', 'Walker2d']
 
 
 def load_baselines(base_dir, verbose=False):
@@ -61,14 +62,14 @@ def load_baselines(base_dir, verbose=False):
   for game in ALL_GAMES:
     for agent in ['dqn', 'c51', 'rainbow', 'iqn']:
       game_data_file = os.path.join(base_dir, agent, '{}.pkl'.format(game))
-      if not tf.gfile.Exists(game_data_file):
+      if not tf.io.gfile.exists(game_data_file):
         if verbose:
           # pylint: disable=superfluous-parens
           print('Unable to load data for agent {} on game {}'.format(agent,
                                                                      game))
           # pylint: enable=superfluous-parens
         continue
-      with tf.gfile.Open(game_data_file, 'rb') as f:
+      with tf.io.gfile.GFile(game_data_file, 'rb') as f:
         if sys.version_info.major >= 3:
           # pylint: disable=unexpected-keyword-arg
           single_agent_data = pickle.load(f, encoding='latin1')
@@ -121,7 +122,7 @@ def load_statistics(log_path, iteration_number=None, verbose=True):
     print('Reading statistics from: {}'.format(log_file))
     # pylint: enable=superfluous-parens
 
-  with tf.gfile.Open(log_file, 'rb') as f:
+  with tf.io.gfile.GFile(log_file, 'rb') as f:
     return pickle.load(f), iteration_number
 
 
@@ -154,7 +155,7 @@ def get_latest_iteration(path):
     ValueError: if there is not available log data at the given path.
   """
   glob = os.path.join(path, '{}_[0-9]*'.format(FILE_PREFIX))
-  log_files = tf.gfile.Glob(glob)
+  log_files = tf.io.gfile.glob(glob)
 
   if not log_files:
     raise ValueError('No log data found at {}'.format(path))
