@@ -386,7 +386,7 @@ class JaxImplicitQuantileAgent(dqn_agent.JaxDQNAgent):
 
     self._rng, self.action = select_action(self.network_def,
                                            self.online_params,
-                                           self.state,
+                                           self.preprocess_fn(self.state),
                                            self._rng,
                                            self.num_quantile_samples,
                                            self.num_actions,
@@ -419,9 +419,9 @@ class JaxImplicitQuantileAgent(dqn_agent.JaxDQNAgent):
             self.target_network_params,
             self.optimizer,
             self.optimizer_state,
-            self.replay_elements['state'],
+            self.preprocess_fn(self.replay_elements['state']),
             self.replay_elements['action'],
-            self.replay_elements['next_state'],
+            self.preprocess_fn(self.replay_elements['next_state']),
             self.replay_elements['reward'],
             self.replay_elements['terminal'],
             self.num_tau_samples,
