@@ -352,7 +352,7 @@ class JaxRainbowAgent(dqn_agent.JaxDQNAgent):
 
     self._rng, self.action = select_action(self.network_def,
                                            self.online_params,
-                                           self.state,
+                                           self.preprocess_fn(self.state),
                                            self._rng,
                                            self.num_actions,
                                            self.eval_mode,
@@ -398,9 +398,9 @@ class JaxRainbowAgent(dqn_agent.JaxDQNAgent):
             self.target_network_params,
             self.optimizer,
             self.optimizer_state,
-            self.replay_elements['state'],
+            self.preprocess_fn(self.replay_elements['state']),
             self.replay_elements['action'],
-            self.replay_elements['next_state'],
+            self.preprocess_fn(self.replay_elements['next_state']),
             self.replay_elements['reward'],
             self.replay_elements['terminal'],
             loss_weights,
